# https://ossf.github.io/osv-schema/


import re
from typing import List


class OSVParseException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class OSVSeverity:
    def __init__(self, type: str, score: str) -> None:
        self.type: str = type
        self.score: str = score


class OSVReference:
    def __init__(self, type: str, url: str) -> None:
        self.type: str = type
        self.url: str = url


class OSVEvent:
    def __init__(self, **kwargs) -> None:
        self.__allowed = ["introduced", "fixed", "last_affected", "limit"]
        """
        Defines a package event.
        Can only one of either introduced, fixed, last_affected or limit.
        """
        [(name, value)] = list(kwargs.items())
        assert name in self.__allowed
        setattr(self, name, value)
        self.__validate()

    def __validate(self):
        pass


class OSVRange:
    def __init__(self, type: str, events: List[OSVEvent], repo: str = None) -> None:
        self.type: str = type
        self.repo: str = repo
        self.events: List[OSVEvent] = events
        self.database_specific: str = None
        self.ecosystem_specific: dict = {}
        self.__validate()

    def __validate(self):
        if self.type == "git" and not self.repo:
            raise OSVParseException("repo is required when type is 'git'")
        if (
            not self.events
            or not len([e for e in self.events if hasattr(e, "introduced")]) >= 1
        ):
            raise OSVParseException(
                "events must have atleast one occurence of an 'introduced' event"
            )


class OSVPackage:
    def __init__(self, ecosystem: str, name: str, package_url: str = None) -> None:
        self.ecosystem: str = ecosystem
        self.name: str = name
        self.package_url: str = package_url
        self.__validate()

    def __validate(self):
        valid_ecosystems = [
            "go",
            "npm",
            "oss-fuzz",
            "pypi",
            "rubygems",
            "crates.io",
            "packagist",
            "maven",
            "nuget",
            "linux",
            "debian",
            "alpine",
            "hex",
            "android",
            "github actions",
            "pub",
        ]
        if self.ecosystem.lower() not in valid_ecosystems:
            raise OSVParseException(
                "{} is not a valid ecosystem".format(self.ecosystem)
            )


class OSVAffected:
    def __init__(
        self, package: OSVPackage, ranges: List[OSVRange], versions: List[str] = None
    ) -> None:
        self.package: OSVPackage = package
        self.ranges: List[OSVRange] = ranges
        self.versions: List[str] = versions


class OSVulnerability:
    def __init__(
        self,
        id: str,
        modified: str,
        published: str,
        withdrawn: str,
        summary: str,
        details: str,
        aliases: List[str],
        severity: List[OSVSeverity],
        affected: List[OSVAffected],
        references: List[OSVReference] = [],
        database_specific: dict = {},
        schema_version="1.3.0",
    ) -> None:
        self.schema_version: str = schema_version
        self.id: str = id
        self.modified: str = modified
        self.published: str = published
        self.withdrawn: str = withdrawn
        self.aliases: List[str] = aliases
        self.related: List[str] = []
        self.summary: str = summary
        self.details: str = details
        self.severity: List[OSVSeverity] = severity
        self.affected: List[OSVAffected] = affected
        self.references: List[OSVReference] = references
        self.database_specific: dict = database_specific
        self.__validate()

    def __validate(self):
        valid_id_prefixes = [
            "A-",
            "GO-",
            "OSV-",
            "PYSEC-",
            "RUSTSEC-",
            "GSD-",
            "GHSA-",
            "LBSEC-",
            "DSA-",
            "DLA-",
            "DTSA-",
            "CVE-",
        ]
        if not self.id.startswith(tuple(valid_id_prefixes)):
            raise OSVParseException("not a valid id - {}".format(self.id))

        if not validate_rfc3339_date(self.modified):
            raise OSVParseException("'modified' does not match RFC3339 date format")

        if self.published and not validate_rfc3339_date(self.published):
            raise OSVParseException("'published' does not match RFC3339 date format")

        if self.withdrawn and not validate_rfc3339_date(self.withdrawn):
            raise OSVParseException("'withdrawn' does not match RFC3339 date format")


def validate_rfc3339_date(date_str: str):
    rfc3339 = r"^((?:(\d{4}-\d{2}-\d{2})T(\d{2}:\d{2}:\d{2}(?:\.\d+)?))(Z|[\+-]\d{2}:\d{2})?)$"
    if re.match(
        rfc3339,
        date_str,
    ):
        return True
    return False



from ._definitions import (
    GitPack,
    GitPackIndex,
    GitTreeItem,
    GitMainIndex,
    GitObjectType,
    GitRepository,
    ObjectCache,
)

__all__ = [
    "GitPack",
    "GitPackIndex",
    "GitTreeItem",
    "GitMainIndex",
    "GitObjectType",
    "GitRepository",
    "ObjectCache",
]



from ._osv import (
    OSVulnerability,
    OSVEvent,
    OSVAffected,
    OSVPackage,
    OSVRange,
    OSVReference,
    OSVSeverity,
    OSVParseException,
)

from ._advisory import Advisory, Package

from ._db import VulnDb
from ._config import VulnDbSource

__all__ = [
    "Advisory",
    "OSVulnerability",
    "OSVEvent",
    "OSVAffected",
    "OSVPackage",
    "OSVRange",
    "OSVReference",
    "OSVSeverity",
    "OSVParseException",
    "Package",
    "VulnDb",
    "VulnDbSource",
]



from enum import IntEnum
import json
import os
from typing import List
import yara
from ludvig.rules import RuleSetSource
import hashlib
from dataclasses import dataclass, field, asdict
from ludvig.vulndb import Advisory, VulnDbSource


class ConfigEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, RuleSetSource):
            return obj.__dict__
        return json.JSONEncoder.default(self, obj)


class Config:
    def __init__(
        self,
        config_path: str,
        rule_sources: list[RuleSetSource] = None,
        vulndb_sources: list[VulnDbSource] = None,
    ) -> None:
        self.config_path = config_path
        self.compiled_rules = os.path.join(config_path, "ludvig.rules")
        self.config_file = os.path.join(config_path, "config.json")
        if rule_sources and "Built-in" not in rule_sources:
            rule_sources.append(
                RuleSetSource(
                    "Built-In",
                    "secrets",
                    "https://github.com/FrodeHus/ludvig-rules/archive/refs/tags/v0.0.1.tar.gz",
                )
            )
        else:
            rule_sources = [
                RuleSetSource(
                    "Built-In",
                    "secrets",
                    "https://github.com/FrodeHus/ludvig-rules/archive/refs/tags/v0.0.1.tar.gz",
                )
            ]

        self.rule_sources = rule_sources
        if not vulndb_sources:
            vulndb_sources = [
                VulnDbSource(
                    "GitHub Advisory",
                    "https://github.com/github/advisory-database/archive/refs/heads/main.zip",
                )
            ]
        self.vulndb_sources = vulndb_sources
        self.vuln_db_file = os.path.join(config_path, "ludvig.db")

    def save(self):
        with open(self.config_file, "w") as f:
            f.write(json.dumps(self.__dict__, indent=4, cls=ConfigEncoder))

    @staticmethod
    def load():
        config_path = os.path.join(os.path.expanduser("~"), ".ludvig")
        if not os.path.exists(config_path):
            os.makedirs(config_path)
        config_file = os.path.join(config_path, "config.json")
        if os.path.exists(config_file):
            with open(config_file, "r") as f:
                config = json.loads(f.read())
                return Config(
                    config["config_path"],
                    Config.parse_rule_sets(config["rule_sources"]),
                )

        return Config(config_path)

    @staticmethod
    def parse_rule_sets(d: dict):
        rules = []
        for rule_source in d:
            rules.append(
                RuleSetSource(
                    rule_source["name"], rule_source["category"], rule_source["url"]
                )
            )


class Severity(IntEnum):
    UNKNOWN = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    CRITICAL = 4


@dataclass
class RuleMatch:
    rule_id: str
    rule_name: str
    severity: Severity = field(default_factory=lambda: Severity.MEDIUM)
    category: str = field(default=None)
    description: str = field(default=None)
    tags: list[str] = field(default_factory=lambda: [])

    @property
    def __dict__(self):
        return asdict(self)

    @staticmethod
    def from_yara_match(yara: yara.Match) -> "RuleMatch":
        return RuleMatch(
            yara.meta["id"] if "id" in yara.meta else "LS00000",
            yara.rule,
            Severity[yara.meta["severity"]]
            if "severity" in yara.meta
            else Severity.UNKNOWN,
            yara.namespace,
            yara.meta["description"] if "description" in yara.meta else "",
            yara.tags,
        )

    def from_vuln_advisory(advisory: Advisory) -> "RuleMatch":
        return RuleMatch(
            advisory.id,
            advisory.ext_id,
            Severity.HIGH,
            advisory.ecosystem,
            advisory.details,
        )


class FindingSample:
    def __init__(
        self, content: str, offset: int, deobfuscated=False, line_number: int = -1
    ) -> None:
        self.offset = offset
        self.line_number = line_number
        content = content[:10] + "..." if len(content) > 10 else content
        if deobfuscated:
            self.content = content
        else:
            obfuscated = "*" * len(content)
            self.content = (
                obfuscated[:10] + "..." if len(obfuscated) > 10 else obfuscated
            )

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)

    @classmethod
    def from_yara_match(
        cls, match: yara.Match, deobfuscated=False, line_number: int = -1
    ) -> List["FindingSample"]:
        samples = []
        for str_match in match.strings:
            offset = str_match[0]
            data = str_match[2]
            if data.isascii():
                data = data.decode("utf-8")
            else:
                data = "".join(format(x, "02x") for x in data)
            samples.append(FindingSample(data, offset, deobfuscated, line_number))
        return samples


@dataclass
class Finding:
    id: str
    category: str
    rule: RuleMatch
    filename: str
    severity: Severity = field(init=False)
    samples: list[FindingSample] = field(default_factory=lambda: [])
    properties: dict = field(default_factory=dict)
    _hash: str = field(init=False, repr=False)

    def __post_init__(self):
        self.name = f"{self.category}/{self.rule.rule_name}"
        self.severity = self.rule.severity
        self.properties.update({"category": self.category})
        self._hash = hashlib.sha1(
            "|".join(
                [
                    self.name,
                    self.filename,
                    "@".join([s.content for s in self.samples]),
                ]
            ).encode()
        ).hexdigest()

    @property
    def __dict__(self):
        return asdict(self)

    @staticmethod
    def from_secret(
        yara_match: yara.Match,
        samples: list[FindingSample],
        file_name: str,
        meta: dict = {},
    ) -> "Finding":
        rule = RuleMatch.from_yara_match(yara_match)
        return Finding(rule.rule_id, rule.category, rule, file_name, samples, meta)

    @staticmethod
    def from_vuln_advisory(
        advisory: Advisory, actual_version: str, filename: str, meta: dict = {}
    ) -> "Finding":
        rule = RuleMatch.from_vuln_advisory(advisory)
        return Finding(
            advisory.id,
            "vulnerabilities",
            rule=rule,
            filename=advisory.package.name,
            properties=meta,
        )



import binascii
import enum
import glob
import hashlib
from io import BufferedReader
import os
import struct
from typing import List
import zlib
from knack.log import get_logger

logger = get_logger(__name__)

byte_order_network = "!"
byte_order_little_endian = "<"

# handy docs:
# - https://shafiul.github.io//gitbook/7_the_packfile.html
# - https://git-scm.com/docs/pack-format, https://codewords.recurse.com/issues/three/unpacking-git-packfiles
# - https://alibaba-cloud.medium.com/a-detailed-explanation-of-the-underlying-data-structures-and-principles-of-git-19914c7906c2


class GitObjectType(enum.Enum):
    NOT_SUPPORTED = 0
    OBJ_COMMIT = 1
    OBJ_TREE = 2
    OBJ_BLOB = 3
    OBJ_TAG = 4
    OBJ_OFS_DELTA = 6
    OBJ_REF_DELTA = 7


class GitDelta:
    def __init__(
        self, target_offset: int, target_size: int, source_offset: int, data
    ) -> None:
        self.target_offset = target_offset
        self.target_size = target_size
        self.source_offset = source_offset
        self.data = data

    def has_data(self):
        """
        true if this delta has data to add to the target
        """
        return self.data is not None


def read(f: BufferedReader, format, byte_order="!"):
    format = "{}{}".format(byte_order, format)
    bytes = f.read(struct.calcsize(format))
    return struct.unpack(format, bytes)[0]


class GitPackIndex(dict):
    def __init__(self, idx_file: str):
        idx = self.__read(idx_file)
        super().__init__(idx)

    def get_offset(self, object_hash: str):
        t = self.__find_index(self["objects"], object_hash)
        obj = self["objects"][t]
        if obj:
            return obj["offset"]
        return None

    def __find_index(self, elements: dict, value):
        value = int("0x" + value, 0)
        left, right = 0, len(elements) - 1
        while left <= right:
            middle = (left + right) // 2
            middle_element = int("0x" + elements[middle]["name"], 0)
            if middle_element == value:
                return middle

            if middle_element < value:
                left = middle + 1
            elif middle_element > value:
                right = middle - 1

    def __read(self, filename: str):
        idx = {}
        with open(filename, "rb") as f:
            signature = f.read(4)
            if signature != b"\xfftOc":
                raise Exception("Not a Git pack index file: %s", filename)
            idx["version"] = read(f, "I")
            fan_out_table = {}
            for n in range(256):
                fan_out_table["%0.2X" % n] = read(f, "I")
            idx["total_objects"] = fan_out_table["FF"]
            idx["objects"] = []
            for n in range(idx["total_objects"]):
                object_name = binascii.hexlify(f.read(20)).decode("ascii")
                idx["objects"].append({"name": object_name, "offset": 0})
            for n in range(idx["total_objects"]):
                crc = read(f, "I")
            for n in range(idx["total_objects"]):
                offset = read(f, "I")
                idx["objects"][n]["offset"] = offset
        return idx


class GitPack:
    def __init__(self, filename: str, idx: dict) -> None:
        self.filename = filename
        self.__fp = open(filename, "rb")
        self.idx = idx
        self.commits = self.__get_all_commits()
        self.entries = self.__read_git_pack()
        self.__cache = {}

    def get_offset_by_hash(self, hash: str):
        return self.idx.get_offset(hash)

    def get_commit_by_hash(self, commit_hash: str):
        commit = [c for c in self.commits if c.hash == commit_hash]
        if len(commit):
            return commit[0]
        return None

    def get_object(
        self,
        hash: str = None,
        offset: int = None,
        meta_data_only=False,
        expected_type: GitObjectType = None,
    ):
        if hash:
            offset = self.get_offset_by_hash(hash)
        if not offset:
            return None

        hash_key = hashlib.sha1(str(offset).encode()).hexdigest()
        if hash_key in self.__cache:
            return self.__cache[hash_key]

        self.__fp.seek(offset, 0)
        (obj_type, size) = self.__read_pack_object_header(self.__fp)
        if meta_data_only:
            return (offset, obj_type, size)
        content = None
        if obj_type == GitObjectType.NOT_SUPPORTED:
            return
        if obj_type == GitObjectType.OBJ_COMMIT:
            content = read_compressed_object(self.__fp, size)
            return parse_commit_message(content, hash)
        elif obj_type == GitObjectType.OBJ_BLOB:
            content = read_compressed_object(self.__fp, size)
        elif obj_type == GitObjectType.OBJ_TREE:
            content = read_compressed_object(self.__fp, size)
        elif obj_type == GitObjectType.OBJ_REF_DELTA:
            object_name = binascii.hexlify(self.__fp.read(20)).decode("ascii")
            content = self.get_object(hash=object_name)
        elif obj_type == GitObjectType.OBJ_OFS_DELTA:
            delta_offset = self.__read_delta_offset(self.__fp)
            content = read_compressed_object(self.__fp, size)
            content, obj_type = self.__parse_delta(
                content, delta_offset, offset, expected_type=expected_type
            )
        if len(content) < 10000:
            self.__cache[hash_key] = (content, obj_type, len(content))
        return content, obj_type, len(content)

    def __parse_delta(
        self,
        delta_data,
        base_object_offset: int,
        current_offset: int,
        expected_type: GitObjectType = None,
    ):
        delta_list = self.__parse_delta_instructions(delta_data)
        if expected_type == GitObjectType.OBJ_BLOB and len(delta_list) > 2:
            crude_delta = "".join(
                [d.data.decode("utf-8", errors="replace") for d in delta_list if d.data]
            )
            if len(crude_delta) > 0:
                return crude_delta.encode(), expected_type
        base_obj, obj_type, size = self.get_object(
            offset=current_offset - base_object_offset
        )

        if obj_type == GitObjectType.OBJ_OFS_DELTA:
            delta_offset = self.__read_delta_offset(self.__fp)
            content = read_compressed_object(self.__fp, size)
            return self.__parse_delta(
                content, delta_offset, current_offset - base_object_offset
            )
        else:
            assembled_object = self.__apply_delta_list(base_obj, delta_list)
            return assembled_object, obj_type

    def get_all_blob_offsets(self):
        for obj in self.idx["objects"]:
            (offset, obj_type, _) = self.get_object(
                hash=obj["name"], meta_data_only=True
            )
            if obj_type == GitObjectType.OBJ_BLOB:
                yield obj["name"], offset

    def object_exists(self, hash: str):
        return len([obj for obj in self.idx["objects"] if obj["name"] == hash])

    def __search_tree(self, tree: "GitTree", match_hash: str):
        for leaf in tree.leafs:
            if leaf.hash == match_hash:
                return leaf
            if leaf.mode == 40000 or leaf.mode == 160000:
                tree = self.get_object(hash=leaf.hash)
                return self.__search_tree(tree, match_hash)
        return None

    def __apply_delta_list(self, source, delta_list: List[GitDelta] = []):
        if len(delta_list) == 0:
            return source
        deltas_applied = b""
        for delta in delta_list:
            deltas_applied = self.__apply_delta(source, deltas_applied, delta)
        return deltas_applied

    def __apply_delta(self, source, data: bytes, delta: GitDelta):
        if delta.has_data():
            return data + delta.data
        else:
            return (
                data
                + source[delta.source_offset : delta.source_offset + delta.target_size]
            )

    def __get_all_commits(self):
        commits = []
        for obj in self.idx["objects"]:
            self.__fp.seek(obj["offset"])
            (obj_type, size) = self.__read_pack_object_header(self.__fp)
            if obj_type == GitObjectType.OBJ_COMMIT:
                content = read_compressed_object(self.__fp, size)
                commit = parse_commit_message(content, obj["name"])
                commits.append(commit)
        return commits

    def __parse_delta_instructions(self, data):
        i, source_length = self.__msb_size(data)
        i, target_length = self.__msb_size(data, i)
        delta_list = []
        target_bytes = 0
        while i < len(data):
            c = data[i]
            i += 1
            if c & 0x80:  # MSB is set - copy instruction
                cp_off, cp_size = 0, 0
                if c & 0x01:
                    cp_off = data[i]
                    i += 1
                if c & 0x02:
                    cp_off |= data[i] << 8
                    i += 1
                if c & 0x04:
                    cp_off |= data[i] << 16
                    i += 1
                if c & 0x08:
                    cp_off |= data[i] << 24
                    i += 1
                if c & 0x10:
                    cp_size = data[i]
                    i += 1
                if c & 0x20:
                    cp_size |= data[i] << 8
                    i += 1
                if c & 0x40:
                    cp_size |= data[i] << 16
                    i += 1

                if not cp_size:
                    cp_size = 0x10000

                rbound = cp_off + cp_size
                if rbound < cp_size or rbound > source_length:
                    break

                delta = GitDelta(target_bytes, cp_size, cp_off, None)
                delta_list.append(delta)
                target_bytes += cp_size
            elif c:  # insert instruction
                delta = GitDelta(target_bytes, c, 0, data[i : i + c])
                delta_list.append(delta)
                i += c
                target_bytes += c
        return delta_list

    def __msb_size(self, data, offset=0):
        """
        Reads variable-length size from byte array
        """
        size = 0
        i = 0
        l = len(data)
        while i < l:
            c = data[i + offset]
            size |= (c & 0x7F) << i * 7
            i += 1
            if not c & 0x80:
                break
        return i + offset, size

    def __read_pack_object_header(self, f: BufferedReader):
        byte0 = read(f, "B")
        obj_type = self.__get_object_type(byte0)
        size = byte0 & 15  # starting size
        s = 4  # starting bit-shift size
        while byte0 & 0x80:  # MSB is 1 so we keep reading to get full length
            byte0 = read(f, "B")
            size += (byte0 & 0x7F) << s
            s += 7
        size += (byte0 & 0x7F) << s
        return (obj_type, size)

    def __read_delta_offset(self, f: BufferedReader):
        """
        Reads variable-length negative offset value (for finding next ofs_delta)
        """
        byte0 = read(f, "B")
        delta_offset = byte0 & 0x7F
        while byte0 & 0x80:
            byte0 = read(f, "B")
            delta_offset += 1
            delta_offset = (delta_offset << 7) + (byte0 & 0x7F)

        return delta_offset

    def __read_git_pack(self):
        pack = {}
        self.__fp.seek(0)
        signature = self.__fp.read(4).decode("ascii")
        if signature != "PACK":
            raise Exception("Not a Git pack file: %s", self.filename)
        pack["version"] = read(self.__fp, "I")
        num_entries = read(self.__fp, "I")
        pack["total_objects"] = num_entries
        for n in range(num_entries):
            pass
        return pack

    def __get_object_type(self, type: int):
        type_id = (type & 0x70) >> 4
        if not type_id in range(1, 8):
            return GitObjectType.NOT_SUPPORTED
        if type_id == 1:
            return GitObjectType.OBJ_COMMIT
        elif type_id == 2:
            return GitObjectType.OBJ_TREE
        elif type_id == 3:
            return GitObjectType.OBJ_BLOB
        elif type_id == 4:
            return GitObjectType.OBJ_TAG
        elif type_id == 6:
            return GitObjectType.OBJ_OFS_DELTA
        elif type_id == 7:
            return GitObjectType.OBJ_REF_DELTA
        return GitObjectType.NOT_SUPPORTED

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        if self.__fp:
            self.__fp.close()


class GitTreeItem:
    def __init__(self, path: str, mode: int, hash: str) -> None:
        self.path = path
        self.mode = mode
        self.hash = hash

    def is_regular_file(self):
        if self.mode & 0x8000:
            return True
        return False

    def is_directory(self):
        if self.mode & 0x4000:
            return True
        return False


class GitTree:
    def __init__(self, items=List[GitTreeItem]) -> None:
        self.leafs = items


class GitCommit:
    def __init__(
        self,
        parent_hash: str,
        tree_hash: str,
        author: str,
        committer: str,
        hash: str = None,
    ) -> None:
        self.hash = hash
        self.parent_hash = parent_hash
        self.tree_hash = tree_hash
        self.auth = author
        self.comitter = committer


class GitRepository:
    def __init__(self, repo_path: str) -> None:
        self.path = repo_path
        self.__packs: List[GitPack] = []
        self.commits = []
        self.__entries = {}
        total_size = 0
        with open(os.path.join(self.path, "HEAD"), "r") as f:
            self.__head = f.read()
            self.__head = self.__head.split()[1]
        refs = self.__read_refs()
        self.__ref_sha = refs[self.__head]
        current_commit, type = self.__get_local_object(self.__ref_sha)
        if type == GitObjectType.OBJ_COMMIT:
            self.commits.append(current_commit)
        else:
            logger.error("expected HEAD to point to a commit object, not %s", type.name)
        pack_files = self.__get_pack_files()
        for pf in pack_files:
            total_size += os.stat(pf + ".pack").st_size / (1024 * 1024)
            idx = GitPackIndex(pf + ".idx")
            pack = GitPack(pf + ".pack", idx)
            self.commits.extend(pack.commits)
            self.__packs.append(pack)
        if total_size > 100:
            logger.warn(
                "Git index is larger than 100 (%dMB) - scanning may be slow [commits: %d]",
                total_size,
                len(self.commits),
            )

    def object_exists(self, hash: str):
        for pack in self.__packs:
            if pack.object_exists(hash):
                return True
        return False

    def get_tree(self, hash: str = None, offset: str = None):
        local_tree, _ = self.__get_local_object(hash)
        if local_tree:
            return local_tree

        content, obj_type, _ = self.get_object(hash, offset)
        if obj_type != GitObjectType.OBJ_TREE:
            logger.debug(
                "tree object %s returned %s - nuked from history?",
                hash or offset,
                obj_type.name,
            )
            return None
        return parse_tree(content)

    def get_object(
        self,
        hash: str = None,
        offset: str = None,
        meta_data_only=False,
        expected_type: GitObjectType = None,
    ):
        local_obj, obj_type = self.__get_local_object(hash)
        if local_obj:
            return local_obj, obj_type, len(local_obj)

        for pack in self.__packs:
            try:
                found, obj_type, size = pack.get_object(
                    hash=hash,
                    offset=offset,
                    meta_data_only=meta_data_only,
                    expected_type=expected_type,
                )
                if found:
                    return found, obj_type, size
            except TypeError as ex:
                logger.debug("could not retrieve %s - not a valid object?", hash)
                logger.debug(ex)
        return None, GitObjectType.NOT_SUPPORTED, -1

    def walk_tree(
        self, tree: "GitTree", path_prefix="", obj_cache: "ObjectCache" = None
    ):
        files = []
        if not hasattr(tree, "leafs"):
            logger.warn("tree has no leafs")
            return files
        for leaf in tree.leafs:
            if leaf.is_regular_file():
                if obj_cache:
                    added, modified = obj_cache.add(leaf, path_prefix)
                    if not added and not modified:
                        continue
                leaf.path = os.path.join(path_prefix, leaf.path)
                files.append(leaf)
            else:
                next_tree = self.get_tree(hash=leaf.hash)
                if not next_tree:
                    logger.debug(
                        "tree %s [%s] was requested but not found in index - nuked?",
                        leaf.hash,
                        os.path.join(path_prefix, leaf.path),
                    )
                    continue
                files.extend(
                    self.walk_tree(
                        next_tree,
                        os.path.join(path_prefix, leaf.path),
                        obj_cache=obj_cache,
                    )
                )
        return files

    def build_history(self):
        pass

    def __get_local_object(self, hash: str):
        obj_path = os.path.join(self.path, "objects", hash[:2], hash[2:])
        if not os.path.exists(obj_path):
            return None, None

        with open(obj_path, "rb") as f:
            inflated = zlib.decompress(f.read())

        start = inflated.find(b"\x00") + 1
        type = inflated[:start].split()[0].decode("utf-8")
        data = inflated[start:]
        if type == "commit":
            type = GitObjectType.OBJ_COMMIT
            parsed = parse_commit_message(data, hash)
        elif type == "tree":
            type = GitObjectType.OBJ_TREE
            parsed = parse_tree(data)
        elif type == "blob":
            type = GitObjectType.OBJ_BLOB
            parsed = data
        else:
            type = GitObjectType.NOT_SUPPORTED
            parsed = data

        return parsed, type

    def __read_refs(self):
        refs = {}
        info_refs = os.path.join(self.path, "info", "refs")
        if os.path.exists(info_refs):
            with open(info_refs) as f:
                ref_data = f.read()
            for line in ref_data.splitlines():
                l = line.split("\t")
                ref = l[1]
                sha = l[0]
                refs[ref] = sha

        head_refs = os.path.join(self.path, "refs", "heads")
        if os.path.exists(head_refs):
            for file in os.listdir(head_refs):
                with open(os.path.join(head_refs, file)) as f:
                    refs[os.path.join("refs", "heads", file)] = f.read().rstrip("\n")
        return refs

    def __get_pack_files(self):
        pack_path = os.path.join(self.path, "objects")
        pack_files = []
        for file in glob.iglob(
            os.path.join(pack_path, "**/pack-*.pack"), recursive=True
        ):
            pack_name = file[:-5]
            pack_files.append(pack_name)
        return pack_files

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        for pack in self.__packs:
            pack.close()


class GitMainIndex(dict):
    def __init__(self, idx_file: str):
        idx = self.__read_git_main_index(idx_file)
        super().__init__(idx)

    def get_files(self):
        for entry in self["entries"]:
            path = os.path.join(
                self.__base_path, "objects", entry["sha1"][:2], entry["sha1"][2:]
            )
            if os.path.exists(path):
                with open(path, "rb") as f:
                    inflated = zlib.decompress(f.read())
                    idx = inflated.find(b"\x00")
                    type = inflated[:idx]
                    content = inflated[idx + 1 :]
                    yield content, entry["name"], entry["sha1"], self.__ref_sha

    def __read_git_main_index(self, index_path: str):
        # docs: https://git-scm.com/docs/index-format
        index = {}

        with open(index_path, "rb") as f:
            signature = f.read(4).decode("ascii")
            if signature != "DIRC":
                raise Exception("Not a Git index file: %s", index_path)

            index["version"] = read(f, "I")
            index["entries"] = []
            num_entries = read(f, "I")
            for i in range(num_entries):
                try:
                    entry = {}
                    entry["entry"] = i + 1
                    entry["ctime_seconds"] = read(f, "I")
                    entry["ctime_nanoseconds"] = read(f, "I")
                    entry["mtime_seconds"] = read(f, "I")
                    entry["mtime_nanoseconds"] = read(f, "I")
                    entry["dev"] = read(f, "I")
                    entry["ino"] = read(f, "I")
                    entry["mode"] = read(f, "I")
                    entry["uid"] = read(f, "I")
                    entry["gid"] = read(f, "I")
                    entry["size"] = read(f, "I")
                    entry["sha1"] = binascii.hexlify(f.read(20)).decode("ascii")
                    entry["flags"] = read(f, "H")
                    file_name_length = entry["flags"] & 0xFFF
                    bytes_read = 62
                    if file_name_length < 0xFFF:
                        entry["name"] = f.read(file_name_length).decode(
                            "utf-8", "replace"
                        )
                        bytes_read += file_name_length
                    else:
                        name = []
                        while True:
                            byte = f.read(1)
                            if byte == "\x00":
                                break
                            name.append(byte)
                        entry["name"] = b"".join(name).decode("utf-8", "replace")
                        bytes_read += 1
                    padlen = (8 - (bytes_read % 8)) or 8
                    _ = f.read(padlen)
                    index["entries"].append(entry)
                except Exception as ex:
                    logger.error(ex)
        return index


class ObjectCache:
    def __init__(self) -> None:
        self.__cache = {}

    def add(self, leaf: GitTreeItem, parent_path: str = None):
        modified_hash = False
        new = False
        if parent_path:
            path = os.path.join(parent_path, leaf.path)
        else:
            path = leaf.path
        if path in self.__cache and self.__cache[path] != leaf.hash:
            new = False
            modified_hash = True
        elif path not in self.__cache:
            new = True
        self.__cache[path] = leaf.hash
        return new, modified_hash


def read_commit(data: BufferedReader, hash: str = None):
    inflated = read_compressed_object(data)
    return parse_commit_message(inflated)


def parse_commit_message(data, hash: str = None):
    info = {"hash": hash}
    msg = data.decode("utf-8").split("\n")
    for i, line in enumerate(msg):
        if line.startswith("tree"):
            info["tree"] = line.split()[1]
        elif line.startswith("parent"):
            info["parent"] = line.split()[1]
        elif line.startswith("author"):
            info["author"] = " ".join(line.split()[1:-2])
        elif line.startswith("committer"):
            info["comitter"] = " ".join(line.split()[1:-2])
    return GitCommit(
        info["parent"] if "parent" in info else None,
        info["tree"],
        info["author"],
        info["comitter"],
        info["hash"],
    )


def read_compressed_object(f: BufferedReader, size: int):
    try:
        content = zlib.decompress(f.read(size))
        return content
    except:
        return None


def parse_tree(content) -> GitTree:
    tree = []
    i = 0
    while i < len(content):
        try:
            x = content.find(b" ", i)
            if x == -1:
                i += 1
                continue
            mode = int(content[i:x], 8)
            i = x + 1
            x = content.find(b"\x00", x)
            path = content[i:x].decode("utf-8")
            i += (x - i) + 1
            x = i + 20
            sha = binascii.hexlify(content[i:x]).decode("ascii")
            i = x
            tree_item = GitTreeItem(path, mode, sha)
            tree.append(tree_item)
        except Exception as ex:
            logger.warn("failed to parse git tree: %s", ex)
            return None
    return GitTree(tree)



import pprint
import socket
import tempfile
from typing import IO
import requests
from urllib3.connection import HTTPConnection
from urllib3.connectionpool import HTTPConnectionPool
from requests.adapters import HTTPAdapter


class DockerConnection(HTTPConnection):
    def __init__(self):
        super().__init__("localhost")

    def connect(self):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect("/var/run/docker.sock")


class DockerConnectionPool(HTTPConnectionPool):
    def __init__(self):
        super().__init__("localhost")

    def _new_conn(self):
        return DockerConnection()


class DockerAdapter(HTTPAdapter):
    def get_connection(self, url, proxies=None):
        return DockerConnectionPool()


class DockerClient:
    def __init__(self) -> None:
        self.__session = self.__get_docker_session()

    def get_version(self):
        data = self.__session.get("http://docker/version").json()
        return data["ApiVersion"]

    def get_image_list(self):
        image_list = self.__session.get("http://docker/images/json").json()
        for image in image_list:
            repo_tags = image["RepoTags"]
            pprint.pprint(repo_tags)

    def inspect_image(self, name: str):
        data = self.__session.get("http://docker/images/{}/json".format(name)).json()
        return data

    def image_history(self, name: str):
        data = self.__session.get("http://docker/images/{}/history".format(name)).json()
        return data

    def export_image(self, name: str) -> IO[bytes]:
        result = self.__session.get("http://docker/images/{}/get".format(name))
        fp = tempfile.TemporaryFile()
        fp.write(result.content)
        fp.flush()
        fp.seek(0)
        return fp

    def __get_docker_session(self) -> requests.Session:
        session = requests.Session()
        session.mount("http://docker", DockerAdapter())
        return session



from typing import IO, List
from ludvig.sbom.dotnet.dependencies import DotnetDependenciesParser
from cyclonedx.model.bom import Bom
from ludvig._types import Finding, Severity
from ._common import BaseScanner
from ludvig.vulndb import VulnDb


class VulnerabilityScanner(BaseScanner):
    def __init__(self) -> None:
        super().__init__(False)
        self.__db = VulnDb()

    def scan_file_data(
        self, file_data: IO[bytes], file_name: str, severity_level: Severity, **kwargs
    ) -> List[Finding]:
        parser = DotnetDependenciesParser(dependency_content=file_data.read())
        bom = Bom.from_parser(parser)
        findings = []
        for component in bom.components:
            results = self.__db.query_advisory(
                package_name=component.name,
                ecosystem=component.purl.type,
                version=component.version,
            )
            for result in results:
                findings.append(
                    Finding.from_vuln_advisory(result, component.version, file_name)
                )
        return findings

    def accepted_files(self) -> List[str]:
        return ["*.deps.json"]



from collections import OrderedDict
import sys
from ludvig import Severity
from knack import CLI, ArgumentsContext, CLICommandsLoader
from knack.commands import CommandGroup
from knack.invocation import CommandInvoker
import ludvig._help  # noqa
from ludvig._format import transform_finding_list, transform_git_finding_list


class LudvigCommandsLoader(CLICommandsLoader):
    def load_command_table(self, args):
        with CommandGroup(
            self, "image", "ludvig.commands.image#{}", help="Container image operations"
        ) as g:
            g.command("scan", "scan", table_transformer=transform_finding_list)
        with CommandGroup(
            self, "fs", "ludvig.commands.filesystem#{}", help="File system operations"
        ) as g:
            g.command("scan", "scan", table_transformer=transform_finding_list)
        with CommandGroup(self, "git", "ludvig.commands.git#{}") as g:
            g.command("scan", "scan", table_transformer=transform_git_finding_list)
        with CommandGroup(self, "rules", "ludvig.commands.rules#{}") as g:
            g.command("download", "download")
            g.command("add repo", "add_repo")
        with CommandGroup(self, "vulndb", "ludvig.commands.vulndb#{}") as g:
            g.command("build", "build")
        return OrderedDict(self.command_table)

    def load_arguments(self, command):
        with ArgumentsContext(self, "image") as ac:
            ac.argument(
                "severity_level", choices=[e.name for e in Severity], default="MEDIUM"
            )
            ac.argument("max_file_size", type=int)
            ac.argument(
                "enabled", choices=["secret", "vuln"], default="secret vuln", nargs="+"
            )
        with ArgumentsContext(self, "fs") as ac:
            ac.argument(
                "severity_level", choices=[e.name for e in Severity], default="MEDIUM"
            )
            ac.argument("max_file_size", type=int)
            ac.argument(
                "enabled", choices=["secret", "vuln"], default="secret vuln", nargs="+"
            )
        with ArgumentsContext(self, "git") as ac:
            ac.argument(
                "severity_level", choices=[e.name for e in Severity], default="MEDIUM"
            )
            ac.argument("max_file_size", type=int)
            ac.argument(
                "enabled", choices=["secret", "vuln"], default="secret vuln", nargs="+"
            )
        super(LudvigCommandsLoader, self).load_arguments(command)


class LudvigCommandInvoker(CommandInvoker):
    def execute(self, args):
        result = super().execute(args)
        if (
            len(args) > 1 and args[1] == "scan" and result.result
        ):  # dirty hack to provoke exit code if any scan results return
            result.exit_code = 1
        return result


ludvig_cli = CLI(
    cli_name="ludvig",
    commands_loader_cls=LudvigCommandsLoader,
    invocation_cls=LudvigCommandInvoker,
)
exit_code = ludvig_cli.invoke(sys.argv[1:])
sys.exit(exit_code)



from ludvig.scanners._common import BaseScanner, ScanPipeline
from ludvig.scanners._secrets import SecretScanner
from ludvig.scanners._vulndb import VulnerabilityScanner

__all__ = [
    "BaseScanner",
    "SecretScanner",
    "VulnerabilityScanner",
    "ScanPipeline",
]



from ludvig import Severity
from ludvig.scanners import ScanPipeline, SecretScanner, VulnerabilityScanner
from ludvig.providers import FileSystemProvider


def scan(
    path: str,
    severity_level: Severity = Severity.MEDIUM,
    deobfuscated=False,
    output_sarif=None,
    max_file_size=10000,
):
    """
    Scans a filesystem path
    :param path: Path to scan
    :param severity_level: Set severity level for reporting
    :param deobfuscated: Returns any secrets found in plaintext. Default: False.
    :param output_sarif: Generates SARIF report if filename is specified.
    :param max_file_size: Max file size for scanning (in bytes).
    """
    if isinstance(severity_level, str):
        severity_level = Severity[severity_level]
    provider = FileSystemProvider(path, max_file_size=max_file_size)
    pipeline = ScanPipeline(
        [SecretScanner(deobfuscated), VulnerabilityScanner()], provider, severity_level
    )
    pipeline.scan()
    if output_sarif:
        from ludvig.outputs import SarifConverter

        report = SarifConverter.from_findings(pipeline.get_unique_findings())
        with open(output_sarif, "w") as r:
            r.write(report)
    return pipeline.findings



import glob
import os
from typing import List
from ._providers import BaseFileProvider


class FileSystemProvider(BaseFileProvider):
    def __init__(
        self, path, exclusions: List[str] = None, max_file_size=200000
    ) -> None:
        super().__init__(exclusions=exclusions, max_file_size=max_file_size)
        self.path = path

    def get_files(self):
        for filename in glob.iglob(os.path.join(self.path, "**/*"), recursive=True):
            if (
                os.path.isdir(filename)
                or self.is_excluded(filename)
                or (
                    os.stat(filename).st_size > self.max_file_size
                    and not filename.endswith(".json")
                )
            ):
                continue
            with open(filename, "rb") as f:
                yield f, filename, {}



import sqlite3
from typing import List
from knack.log import get_logger
from ludvig.vulndb import Advisory, Package

logger = get_logger(__name__)


class VulnDb(object):
    __sql_create_advisory_table = """ CREATE TABLE IF NOT EXISTS Advisory (
                                    id integer PRIMARY KEY AUTOINCREMENT,
                                    package_id INTEGER NOT NULL,
                                    modified datetime,
                                    published datetime,
                                    ext_id text NOT NULL COLLATE NOCASE,
                                    ecosystem text NOT NULL COLLATE NOCASE,
                                    summary varchar(200),
                                    details text,
                                    version text NOT NULL,
                                    fixed text,
                                    source varchar(100),
                                    FOREIGN KEY (package_id) REFERENCES Package(id)
                                ); """

    __sql_create_alias_table = """CREATE TABLE IF NOT EXISTS Alias (
                                    id integer PRIMARY KEY AUTOINCREMENT,
                                    advisory_id INTEGER NOT NULL,
                                    alias varchar(200) COLLATE NOCASE,
                                    FOREIGN KEY (advisory_id) REFERENCES Advisory(id)
                                );
    """

    __sql_create_package_table = """CREATE TABLE IF NOT EXISTS Package (
                                        id integer PRIMARY KEY AUTOINCREMENT,
                                        name text NOT NULL COLLATE NOCASE
    )
    """

    def __init__(self, db_file: str = None) -> None:
        if not db_file:
            from ludvig import Config

            current_config = Config.load()
            db_file = current_config.vuln_db_file
        self.__conn = self.__get_connection(db_file)

    def __get_connection(self, db_file: str):

        conn = None
        try:
            conn = sqlite3.connect(db_file)
            if not self.__verify_db_exists(conn.cursor()):
                self.__create_db(conn.cursor())
            return conn
        except sqlite3.Error as e:
            logger.error(e)

    def __verify_db_exists(self, c: sqlite3.Cursor):
        res = c.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name='advisory';"
        )
        result = res.fetchone()
        if result and "advisories" in result:
            return True
        return False

    def __create_db(self, c: sqlite3.Cursor):
        c.execute(self.__sql_create_package_table)
        c.execute(self.__sql_create_advisory_table)
        c.execute(self.__sql_create_alias_table)

    def add_advisories(self, advisories: List[Advisory]):
        c = self.__conn.cursor()
        for advisory in advisories:
            p = c.execute(
                "SELECT id from package where name = ?", (advisory.package.name,)
            ).fetchone()
            if not p:
                c.execute(
                    "INSERT INTO package (name) VALUES(?)", (advisory.package.name,)
                )
                package_id = c.lastrowid
            else:
                package_id = p[0]

            a = self.query_advisory(
                advisory.package.name,
                advisory.ecosystem,
                advisory.affected_version,
            )
            if not a:
                try:
                    c.execute(
                        """INSERT INTO advisory (package_id, ext_id, modified, published, ecosystem, summary, details, version, fixed, source)
                            VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                            """,
                        (
                            package_id,
                            advisory.ext_id,
                            advisory.modified,
                            advisory.published,
                            advisory.ecosystem,
                            advisory.summary,
                            advisory.details,
                            advisory.affected_version,
                            advisory.fixed_version
                            if hasattr(advisory, "fixed_version")
                            else None,
                            advisory.source,
                        ),
                    )
                    advisory_id = c.lastrowid
                    if advisory.aliases:
                        for alias in advisory.aliases:
                            c.execute(
                                "INSERT INTO Alias (advisory_id, alias) VALUES (?, ?)",
                                (
                                    advisory_id,
                                    alias,
                                ),
                            )
                except Exception as e:
                    logger.error(e)
        self.__conn.commit()

    def query_advisory(
        self,
        package_name: str,
        ecosystem: str,
        version: str,
    ):
        cursor = self.__conn.cursor()

        cursor.execute(
            """
            SELECT a.id, a.ext_id, p.name, a.summary, a.details, a.ecosystem, a.version, a.fixed
            FROM Advisory a
            JOIN Package p ON a.package_id = p.id
            WHERE p.name = ? AND a.ecosystem = ? and (a.version = ? or a.version = '0') COLLATE NOCASE
        """,
            (
                package_name,
                ecosystem,
                version,
            ),
        )
        result = cursor.fetchone()
        if result:
            return [
                Advisory(
                    result[0],
                    result[1],
                    None,
                    Package(result[2]),
                    result[3],
                    None,
                    result[4],
                    result[5],
                    None,
                    result[6],
                )
            ]
        return []

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.__conn.close()



import os
from unittest import TestCase
import mock
from ludvig.scanners import VulnerabilityScanner
from ludvig import Severity
from ludvig.vulndb import Advisory, Package
import ludvig.vulndb


class TestVulnDbDotNet(TestCase):
    @mock.patch.object(ludvig.vulndb.VulnDb, "query_advisory", autospec=True)
    def test_can_detect_vulnerable_package_in_project(self, queryMock: mock.Mock):
        deps_file = os.path.join(
            os.path.dirname(__file__), "fixtures/dotnet.deps-example.json"
        )
        return_value = Advisory(
            1,
            "test",
            None,
            Package("Newtonsoft.Json"),
            "nuget",
            None,
            "test advisory",
            "0",
        )

        def side_effect(*args, **kwargs):
            if kwargs["package_name"] == "Newtonsoft.Json":
                return [return_value]
            return []

        queryMock.side_effect = side_effect
        scanner = VulnerabilityScanner()
        with open(deps_file, "r") as f:
            findings = scanner.scan_file_data(f, "deps.json", Severity.MEDIUM)
        self.assertEqual(len(findings), 1)
        self.assertEqual(findings[0].filename, "Newtonsoft.Json")
        queryMock.assert_called_with(
            mock.ANY,
            package_name="Newtonsoft.Json",
            ecosystem="nuget",
            version="13.0.0",
        )



import abc
import fnmatch
from typing import IO, Dict, List
from ludvig._types import Finding, Severity
from ludvig.providers import BaseFileProvider
from knack.log import get_logger

logger = get_logger(__name__)


class BaseScanner(abc.ABC):
    def __init__(self, deobfuscated=False) -> None:
        self.deobfuscated = deobfuscated

    @abc.abstractmethod
    def scan_file_data(
        self, file_data: IO[bytes], file_name: str, severity_level: Severity, **kwargs
    ) -> List[Finding]:
        """Scans the provided file and reports any findings that matches the given severity level.

        Args:
            file_data (IO[bytes]): The actual file content
            file_name (str): Name of the file
            severity_level (Severity): Severity level to report on

        Keyword Arguments:
            Any additional keyword arguments are added as properties to the finding (metadata)

        Returns:
            List[Finding]: Security issues detected
        """
        pass

    @abc.abstractmethod
    def accepted_files(self) -> List[str]:
        pass

    def accepts_file(self, filename: str):
        for accepted in self.accepted_files():
            if fnmatch.fnmatch(filename, accepted):
                return True
        return False


class ScanPipeline:
    def __init__(
        self,
        scanners: List[BaseScanner],
        provider: BaseFileProvider,
        severity_level: Severity = Severity.MEDIUM,
    ) -> None:
        self.__scanners = scanners
        self.__provider = provider
        self.findings: Dict[str, List[Finding]] = {}
        self.__severity_level = severity_level

    def scan(self):
        """Scans every file returned by the given provider using the provided list of scanners."""
        for file_data, filename, properties in self.__provider.get_files():
            for scanner in self.__scanners:
                if not hasattr(scanner, "scan_file_data"):
                    logger.error(
                        "%s does not implement required method 'scan_file_data'",
                        scanner.__class__.__name__,
                    )
                logger.info("scanning using %s", scanner.__class__.__name__)
                if not scanner.accepts_file(filename):
                    continue
                findings = scanner.scan_file_data(
                    file_data, filename, self.__severity_level, **properties
                )
                if hasattr(file_data, "seek") and callable(getattr(file_data, "seek")):
                    file_data.seek(0)
                self.register_findings(findings)
        self.close_scanners()

    def close_scanners(self):
        for scanner in self.__scanners:
            if hasattr(scanner, "close") and callable(getattr(scanner, "close")):
                scanner.close()

    def register_findings(self, findings: List[Finding]):
        unique_hashes = []
        for category in self.findings:
            unique_hashes.extend([f._hash for f in self.findings[category]])

        for finding in findings:
            if finding._hash in unique_hashes:
                continue

            category = finding.properties["category"]
            if category not in self.findings:
                self.findings[category] = []
            self.findings[category].append(finding)



import os
from ludvig.vulndb import (
    OSVulnerability,
    OSVSeverity,
    OSVRange,
    OSVAffected,
    OSVEvent,
    OSVPackage,
    OSVReference,
    OSVParseException,
)
import json
from knack.log import get_logger

logger = get_logger(__name__)


def read_repository(path: str):
    """Reads every GitHub advisory in a directory

    Args:
        path (str): Path to the directory containing advisories
    """
    for root, _, files in os.walk(path):
        for file in files:
            if file.startswith("GHSA-"):
                p = os.path.join(root, file)
                try:
                    yield read_advisory(p)
                except OSVParseException as p:
                    logger.warn(p)


def read_advisory(file: str) -> OSVulnerability:
    with open(file, "r") as f:
        osv_data = json.load(f)
    severity = [
        OSVSeverity(item["type"], item["score"]) for item in osv_data["severity"]
    ]
    affected = []
    ranges = []
    for item in osv_data["affected"]:
        package = OSVPackage(
            item["package"]["ecosystem"],
            item["package"]["name"],
            item["package"]["purl"] if "purl" in item["package"] else None,
        )

        if "ranges" in item:
            for r in item["ranges"]:
                events = []
                for e in r["events"]:
                    if "introduced" in e:
                        events.append(OSVEvent(introduced=e["introduced"]))
                    elif "fixed" in e:
                        events.append(OSVEvent(fixed=e["fixed"]))
                    elif "last_affected" in e:
                        events.append(OSVEvent(last_affected=e["last_affected"]))
                    elif "limit" in e:
                        events.append(OSVEvent(limit=e["limit"]))

                ranges.append(
                    OSVRange(
                        r["type"],
                        repo=r["repo"] if "repo" in r else None,
                        events=events,
                    )
                )
        a = OSVAffected(package=package, ranges=ranges)
        affected.append(a)
    references = []
    if "references" in osv_data:
        for rf in osv_data["references"]:
            references.append(OSVReference(rf["type"], rf["url"]))
    database_specific = (
        osv_data["database_specific"] if "database_specific" in osv_data else {}
    )
    osv = OSVulnerability(
        osv_data["id"],
        osv_data["modified"],
        osv_data["published"] if "published" in osv_data else None,
        osv_data["withdrawn"] if "withdrawn" in osv_data else None,
        summary=osv_data["summary"] if "summary" in osv_data else None,
        details=osv_data["details"] if "details" in osv_data else None,
        aliases=osv_data["aliases"] if "aliases" in osv_data else [],
        severity=severity,
        affected=affected,
        references=references,
        database_specific=database_specific,
    )
    return osv



import abc
from typing import List
import os
import fnmatch


class BaseFileProvider(abc.ABC):
    def __init__(self, exclusions: List[str] = None, max_file_size=200000) -> None:
        self.exclusions = exclusions or []
        self.__read_ignore_file()
        self.max_file_size = max_file_size

    @abc.abstractmethod
    def get_files(self):
        """Retrieves all files and retrieves them for use with scanners.

        The list is filtered by any pattern defined in `.ludvignore` and configured max file size.
        """
        pass

    def is_excluded(self, filename: str) -> bool:
        for exclude in self.exclusions:
            if fnmatch.fnmatch(filename, exclude):
                return True
        return False

    def __read_ignore_file(self):
        if os.path.exists(".ludvignore"):
            with open(".ludvignore", "r") as fp:
                ignore_data = fp.readlines()
                self.exclusions.extend(ignore_data)



from collections import OrderedDict


def transform_finding_list(result):
    transformed = []
    for type in result:
        for r in result[type]:
            res = OrderedDict(
                [
                    ("Id", r["id"]),
                    ("Name", r["rule"]["rule_name"]),
                    ("Severity", r["severity"]),
                    ("Filename", r["filename"]),
                ]
            )
            transformed.append(res)
    return transformed


def transform_git_finding_list(result):
    transformed = []
    for r in result:

        res = OrderedDict(
            [
                ("RuleId", r["rule"]["rule_id"]),
                ("Name", r["rule"]["ruleName"]),
                ("Severity", r["severity"]),
                ("Filename", r["filename"]),
                ("LineNumber", r["samples"][0]["lineNumber"]),
                ("File SHA", r["properties"]["file_hash"]),
                ("Commit SHA", r["properties"]["commit_hash"]),
            ]
        )
        transformed.append(res)
    return transformed



from knack.help_files import helps

helps[
    "image"
] = """
type: group
short-summary: Operations related to container images
"""

helps[
    "fs"
] = """
type: group
short-summary: Operations related to file system
"""

helps[
    "git"
] = """
type: group
short-summary: Operations related to Git repository
"""

helps[
    "rules"
] = """
type: group
short-summary: Maintain rule sets
"""

helps[
    "vulndb"
] = """
type: group
short-summary: Vulnerability database tasks
"""



__version__ = "0.6.0"

from ._types import (
    Config,
    Finding,
    FindingSample,
    RuleMatch,
    Severity,
)

__all__ = [
    "Config",
    "Finding",
    "FindingSample",
    "RuleMatch",
    "Severity",
]



from ._common import BaseScanner
from ludvig.rules import load_yara_rules
from ludvig.rules import download_rules
from ludvig import Config
import os
from knack.log import get_logger
from typing import IO, List
from ludvig import Finding, FindingSample, Severity

logger = get_logger(__name__)


class SecretScanner(BaseScanner):
    def __init__(self, deobfuscated=False) -> None:
        super().__init__(deobfuscated)
        current_config = Config.load()
        if not os.path.exists(current_config.compiled_rules):
            logger.info("no rules found - downloading defaults...")
            download_rules(current_config.rule_sources, current_config.config_path)

        yara_rules = load_yara_rules(current_config.compiled_rules)
        self.__yara_rules = yara_rules

    def scan_file_data(
        self, file_data: IO[bytes], file_name: str, severity_level: Severity, **kwargs
    ) -> List[Finding]:
        findings = []
        try:
            matches = self.__yara_rules.match(data=file_data.read())
            for match in matches:

                severity = (
                    Severity[match.meta["severity"]]
                    if "severity" in match.meta
                    else Severity.UNKNOWN
                )
                if severity < severity_level:
                    continue
                line = self.find_match_line_num(file_data, match)
                samples = FindingSample.from_yara_match(match, self.deobfuscated, line)
                finding = Finding.from_secret(match, samples, file_name, **kwargs)
                findings.append(finding)
        except Exception as ex:
            return print(ex)
        finally:
            return findings

    def find_match_line_num(self, file_data, match):
        file_data.seek(0)
        fd = file_data.read(match.strings[0][0])
        line = fd.count(b"\n") + 1
        return line

    def accepted_files(self) -> List[str]:
        return ["*"]



import enum
from cyclonedx.output import (
    BaseOutput,
    OutputFormat,
    SchemaVersion,
    get_instance as get_output_instance,
)


@enum.unique
class _CLI_OUTPUT_FORMAT(enum.Enum):
    XML = "xml"
    JSON = "json"


_output_formats = {
    _CLI_OUTPUT_FORMAT.XML: OutputFormat.XML,
    _CLI_OUTPUT_FORMAT.JSON: OutputFormat.JSON,
}
_output_default_filenames = {
    _CLI_OUTPUT_FORMAT.XML: "cyclonedx.xml",
    _CLI_OUTPUT_FORMAT.JSON: "cyclonedx.json",
}



import glob
from io import BytesIO
import time
from typing import List
from ._providers import BaseFileProvider
from ._git import GitRepository, ObjectCache, GitObjectType
import os
from knack import log

logger = log.get_logger(__name__)


class GitRepositoryProvider(BaseFileProvider):
    def __init__(
        self,
        path: str,
        commit: str = None,
        fast_scan=False,
        exclusions: List[str] = None,
        max_file_size=10000,
    ) -> None:
        super().__init__(exclusions=exclusions, max_file_size=max_file_size)
        self.path = path
        self.commit = commit
        self.fast_scan = fast_scan

    def get_files(self):
        repos = glob.iglob(os.path.join(self.path, "**/.git"), recursive=True)
        for repo in repos:

            start_time = time.time()
            time_total = 0
            time_commit_avg = 0
            with GitRepository(repo) as repo:
                if self.commit:
                    commits = [
                        commit for commit in repo.commits if commit.hash == self.commit
                    ]
                else:
                    commits = repo.commits

                num_commits = len(commits)
                obj_cache = ObjectCache()
                for idx, commit in enumerate(commits, start=1):
                    time_commit = time.time()
                    try:
                        if time.time() - start_time > 300:
                            logger.warn(
                                "scanning is taking a long time... [status: %d / %d commits | %.2f commits/s | est. completed: %d min]",
                                idx,
                                num_commits,
                                time_commit_avg,
                                (((num_commits - idx) / time_commit_avg)) / 60,
                            )
                            start_time = time.time()
                        tree = repo.get_tree(hash=commit.tree_hash)
                        if not tree:
                            logger.warn("failed to read tree %s", commit.tree_hash)
                            continue
                        for leaf in repo.walk_tree(tree, obj_cache=obj_cache):
                            if self.is_excluded(leaf.path):
                                continue
                            content, _, size = repo.get_object(
                                hash=leaf.hash,
                                expected_type=GitObjectType.OBJ_BLOB
                                if self.fast_scan and leaf.is_regular_file()
                                else None,
                            )
                            if not content or size > self.max_file_size:
                                continue
                            with BytesIO(content) as c:
                                yield c, leaf.path, {
                                    "file_hash": leaf.hash,
                                    "commit_hash": commit.hash,
                                }
                        time_commit_scan = time.time() - time_commit
                        time_total += time_commit_scan
                        time_commit_avg = idx / time_total
                    except Exception as ex:
                        logger.error(ex)
                        continue



import os
import uuid
from ludvig.vulndb.ingesters import github_read_repository
from knack.log import get_logger
from ludvig.vulndb import VulnDb, Package, Advisory
from ludvig import Config
from tempfile import TemporaryDirectory

logger = get_logger(__name__)


def build():
    """
    Builds the vulnerability database from scratch
    """
    import urllib.request
    from zipfile import ZipFile

    current_config = Config.load()
    if os.path.exists(current_config.vuln_db_file):
        os.remove(current_config.vuln_db_file)

    for src in current_config.vulndb_sources:
        with TemporaryDirectory(prefix="ludvig_vulndb_") as p:
            print(f"Downloading {src.name} from {src.uri}...")
            file = os.path.join(p, f"{uuid.uuid1().hex}.zip")
            urllib.request.urlretrieve(src.uri, file)
            print("Extracting...")
            ZipFile(file).extractall(p)
            print("Adding to database...")
            add_osv_repository(p)


def add_osv_repository(path: str):
    """
    Parses OSV advisories and adds to Ludvig's database
    """
    advisories = []
    with VulnDb() as vulnDb:
        for idx, ghsa in enumerate(github_read_repository(path), start=1):
            for affected in ghsa.affected:
                package = Package(affected.package.name)

                advisory = Advisory(
                    id=None,
                    ext_id=ghsa.id,
                    published=ghsa.published,
                    modified=ghsa.modified,
                    aliases=ghsa.aliases,
                    package=package,
                    ecosystem=affected.package.ecosystem,
                    summary=ghsa.summary,
                    details=ghsa.details,
                    affected_version="",
                    source="ghsa",
                )
                for r in affected.ranges:
                    for e in r.events:
                        if hasattr(e, "introduced"):
                            advisory.affected_version = e.introduced
                        elif hasattr(e, "fixed"):
                            advisory.fixed_version = e.fixed
                if advisory.affected_version:
                    advisories.append(advisory)
                if len(advisories) > 100:
                    vulnDb.add_advisories(advisories)
                    advisories.clear()
        vulnDb.add_advisories(advisories)



from tarfile import TarFile
from typing import IO, List


class Layer:
    def __init__(self, id: str, created_by: str = None, empty_layer=False) -> None:
        self.id = id
        self.created_by = created_by
        self.empty_layer = empty_layer


class Image:
    def __init__(
        self,
        repo_tags: List[str],
        layers: List[Layer],
        environment: List[str],
        image_archive: TarFile,
    ) -> None:
        self.repo_tags = repo_tags
        self.layers = layers
        self.image_archive = image_archive
        self.environment = environment

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.image_archive.close()


class LayerFile:
    def __init__(self, file_data: IO[bytes]) -> None:
        self.file_data = file_data

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.file_data:
            self.file_data.close()



import os
from unittest import TestCase
from ludvig.vulndb.ingesters import github_read_advisory, github_read_repository
from ludvig.vulndb import OSVPackage, OSVRange, OSVEvent, OSVulnerability


class TestParseGitHubAdvisory(TestCase):
    def test_parse_advisory(self):
        ghsa_file = os.path.join(
            os.path.dirname(__file__), "fixtures/ghsa/GHSA-abcde-001.json"
        )
        advisory = github_read_advisory(ghsa_file)
        self.assertIsNotNone(advisory)
        self.assertEqual(len(advisory.affected), 1)
        self.assertEqual(len(advisory.affected[0].ranges), 1)
        self.assertEqual(len(advisory.affected[0].ranges[0].events), 2)
        self.assertEqual(advisory.database_specific["github_reviewed"], True)

    def test_read_repository(self):
        repo = os.path.join(os.path.dirname(__file__), "fixtures/ghsa")
        advisories = list(github_read_repository(repo))
        self.assertEqual(len(advisories), 2)


class TestOpenSourceVulnerabilityValidations(TestCase):
    def test_id_prefix_validation(self):
        expected_prefixes = [
            "A",
            "GO",
            "OSV",
            "PYSEC",
            "RUSTSEC",
            "GSD",
            "GHSA",
            "LBSEC",
            "DSA",
            "DLA",
            "DTSA",
            "CVE",
        ]
        for prefix in expected_prefixes:
            id = "%s-test-advisory" % prefix
            with self.subTest(prefix=prefix):
                OSVulnerability(
                    id,
                    "2023-01-01T15:21:32.59+02:00",
                    None,
                    None,
                    None,
                    None,
                    [],
                    [],
                    [],
                )

    def test_ecosystem_validation(self):
        expected_ecosystems = [
            "go",
            "npm",
            "oss-fuzz",
            "pypi",
            "rubygems",
            "crates.io",
            "packagist",
            "maven",
            "nuget",
            "linux",
            "debian",
            "alpine",
            "hex",
            "android",
            "github actions",
            "pub",
        ]
        for ecosystem in expected_ecosystems:
            with self.subTest(ecosystem=ecosystem):
                OSVPackage(ecosystem, "test")

    def test_range_validation(self):
        test_cases = [
            {
                "type": "git",
                "repo": "test/repo",
                "events": [OSVEvent(introduced="2023-01-01T15:21:32.59+02:00")],
            },
            {
                "type": "semver",
                "repo": None,
                "events": [OSVEvent(introduced="2023-01-01T15:21:32.59+02:00")],
            },
            {
                "type": "ecosystem",
                "repo": None,
                "events": [OSVEvent(introduced="2023-01-01T15:21:32.59+02:00")],
            },
        ]
        for case in test_cases:
            with self.subTest(type=case["type"], repo=case["repo"]):
                OSVRange(case["type"], case["events"], repo=case["repo"])



from types import MappingProxyType
from typing import List
from ludvig import Finding, RuleMatch
import sarif_om as sarif
from jschema_to_python.to_json import to_json

level_mapper = MappingProxyType(
    {
        "CRITICAL": "10",
        "HIGH": "8",
        "MEDIUM": "5",
        "LOW": "3",
        "UNKNOWN": "0",
    }
)

ludvig_tool = sarif.Tool(
    driver=sarif.ToolComponent(
        name="Ludvig SARIF Report",
        full_name="Ludvig by Reothor SARIF Report",
        semantic_version="0.2",
        information_uri="https://github.com/frodehus/ludvig",
    )
)


class SarifConverter:
    def from_findings(findings: List[Finding]) -> str:
        results = SarifConverter.findings_to_results(findings)
        ludvig_tool.driver.rules = SarifConverter.rules_from_findings(findings)
        run = sarif.Run(tool=ludvig_tool, results=results)
        return to_json(
            sarif.SarifLog(
                version="2.1.0",
                schema_uri="http://json.schemastore.org/sarif-2.1.0",
                runs=[run],
            )
        )

    @staticmethod
    def to_sarif_rule_id(rule_match: RuleMatch) -> str:
        return rule_match.rule_id

    @staticmethod
    def to_sarif_level(finding: Finding) -> str:
        return level_mapper.get(finding.rule.severity, "0")

    @staticmethod
    def findings_to_results(findings: List[Finding]) -> List[sarif.Result]:
        results = []
        for finding in findings:
            result = sarif.Result(
                rule_id=SarifConverter.to_sarif_rule_id(finding.rule),
                locations=[SarifConverter.to_sarif_location(finding)],
                message=SarifConverter.to_message(finding.rule),
            )
            results.append(result)
        return results

    @staticmethod
    def to_region(finding: Finding) -> sarif.Region:
        return sarif.Region(
            start_line=finding.samples[0].line_number,
        )

    @staticmethod
    def to_sarif_location(finding: Finding) -> sarif.Location:
        return sarif.Location(
            physical_location=sarif.PhysicalLocation(
                artifact_location=sarif.ArtifactLocation(
                    uri=finding.filename, uri_base_id="%SRCROOT%"
                ),
                region=SarifConverter.to_region(finding),
            ),
        )

    @staticmethod
    def to_message(rule_match: RuleMatch) -> sarif.Message:
        if rule_match.rule_id.startswith("LS"):
            return sarif.Message(text="Secret or other sensitive information found")
        elif rule_match.rule_id.startswith("LM"):
            return sarif.Message(text="Potentially malicious code")
        elif rule_match.rule_id.startswith("LV"):
            return sarif.Message(text="Vulnerable package")
        return sarif.Message(text="Unknown finding")

    @staticmethod
    def rules_from_findings(findings: List[Finding]) -> List[sarif.ReportingDescriptor]:
        rules = []
        for finding in findings:
            rule_id = finding.rule.rule_name
            description = finding.rule.description
            rule = sarif.ReportingDescriptor(
                id=SarifConverter.to_sarif_rule_id(finding.rule),
                name=rule_id,
                short_description=sarif.MultiformatMessageString(
                    text=description, markdown=description
                ),
                properties=sarif.PropertyBag(),
            )
            setattr(
                rule.properties,
                "security-severity",
                SarifConverter.to_sarif_level(finding),
            )

            rules.append(rule)
        return rules



from ._containerprovider import ContainerProvider
from ._filesystemprovider import FileSystemProvider
from ._gitprovider import GitRepositoryProvider
from ._providers import BaseFileProvider

__all__ = [
    "ContainerProvider",
    "FileSystemProvider",
    "GitRepositoryProvider",
    "BaseFileProvider",
]



import json
from typing import List


class DotnetLibrary:
    def __init__(
        self, name: str, version: str, path: str, sha512: str, hash_path: str
    ) -> None:
        self.name = name
        self.version = version
        self.path = path
        self.sha512 = sha512
        self.hash_path = hash_path


class DependenciesFile:
    def __init__(self, libraries: List[DotnetLibrary]) -> None:
        self.libraries = libraries

    @classmethod
    def from_string(cls, content: str) -> "DependenciesFile":
        deps = json.loads(content)
        libs = []
        for _, item in enumerate(deps["libraries"]):
            lib_data = deps["libraries"][item]
            if lib_data["type"] != "package":
                continue
            name_version = item.split("/")
            library = DotnetLibrary(
                name_version[0],
                name_version[1],
                lib_data["path"],
                lib_data["sha512"],
                lib_data["hashPath"],
            )
            libs.append(library)

        return DependenciesFile(libs)



from typing import List
from dateutil.parser import parse
from dataclasses import dataclass, asdict, field


@dataclass
class Package:
    name: str

    @property
    def __dict__(self):
        return asdict(self)


@dataclass
class Advisory:
    id: int
    ext_id: str
    published: str
    package: Package
    ecosystem: str
    summary: str
    details: str
    affected_version: str
    modified: str = field(default=None)
    fixed_version: str = field(default=None)
    aliases: List[str] = field(default=None)
    source: str = field(default=None)

    def __post_init__(self):
        self.id = f"LV{self.id:06}" if self.id is not None else self.id
        if self.modified:
            self.modified = parse(self.modified)
        if self.published:
            self.published = parse(self.published)

    @property
    def __dict__(self):
        return asdict(self)



import json
from unittest import TestCase
from ludvig import Finding, FindingSample, RuleMatch, Severity
from ludvig.outputs import SarifConverter


class TestSarifLogGenerator(TestCase):
    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)

    def test_result_from_finding(self) -> None:
        sample = FindingSample("test", 0)
        finding = Finding(
            1,
            "secret",
            RuleMatch(
                "test", "GitHub Token", Severity.CRITICAL, "secrets", "Test rule"
            ),
            "testfile.json",
            [sample],
        )
        report = SarifConverter.from_findings([finding])
        report = json.loads(report)
        self.assertEqual(len(report["runs"][0]["results"]), 1)



import json
import tarfile
from ._definitions import Image, Layer
from ._client import DockerClient


def read_local_docker_image(repository: str) -> Image:
    client = DockerClient()
    file = client.export_image(repository)
    if not tarfile.is_tarfile(file):
        raise Exception("Not an image")
    file.seek(0)
    img = tarfile.open(fileobj=file)
    with img.extractfile("manifest.json") as mf:
        manifest = json.load(mf)
    with img.extractfile(manifest[0]["Config"]) as cf:
        config = json.load(cf)

    file_layers = [layer for layer in config["history"] if "empty_layer" not in layer]
    layers = []
    for idx, layer in enumerate(manifest[0]["Layers"]):
        layer_id = layer[: layer.index("/")]
        layer_history = __get_layer_history(file_layers, idx + 1)
        layers.append(
            Layer(
                layer_id,
                layer_history["created_by"],
                (
                    layer_history["empty_layer"]
                    if "empty_layer" in layer_history
                    else False
                ),
            )
        )
    return Image(manifest[0]["RepoTags"], layers, config["config"]["Env"], img)


def __get_layer_history(config: dict, layer_index: int):
    return config[layer_index - 1]



from unittest import TestCase
import os
from ludvig.sbom.dotnet.dependencies import DotnetDependenciesParser
from cyclonedx.model.bom import Bom
from cyclonedx.output import OutputFormat, get_instance as get_output


class TestDotnetDependenciesParser(TestCase):
    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)

    def test_simple(self) -> None:
        with open(
            os.path.join(os.path.dirname(__file__), "fixtures/dotnet.deps-example.json")
        ) as r:
            parser = DotnetDependenciesParser(dependency_content=r.read())
            bom = Bom.from_parser(parser)
            v = get_output(bom=bom, output_format=OutputFormat.JSON)
